package main

import (
	"bufio"
	"fmt"
	"os"
	"strconv"
	"strings"
	"time"
	"unicode"
)

func main() {
	start := time.Now()
	partOne()
	duration := time.Since(start)
	partTwo()
	duration2 := time.Since(start)
	fmt.Printf("p1: %s, p2: %s\n", duration, duration2-duration)
}

//To try to debug the problem, they have created a list (your puzzle input) of passwords (according to the corrupted database) and the corporate policy when that password was set.
// For example, suppose you have the following list:
// 1-3 a: abcde
// 1-3 b: cdefg
// 2-9 c: ccccccccc
// Each line gives the password policy and then the password. The password policy indicates the lowest and highest number of times a given letter must appear for the password to be valid. For example, 1-3 a means that the password must contain a at least 1 time and at most 3 times.
// In the above example, 2 passwords are valid. The middle password, cdefg, is not; it contains no instances of b, but needs at least 1. The first and third passwords are valid: they contain one a or nine c, both within the limits of their respective policies.
// How many passwords are valid according to their policies?

type entry struct {
	min      int
	max      int
	letter   string
	password string
}

func fn(c rune) bool {
	// this FieldsFunc treats the - and : as just more separators, which they are
	return !unicode.IsLetter(c) && !unicode.IsNumber(c)
}

func partOne() {
	f, _ := os.Open("input")
	reader := bufio.NewReader(f)
	scanner := bufio.NewScanner(reader)

	passwords := []entry{}

	for scanner.Scan() {
		split := strings.FieldsFunc(scanner.Text(), fn)
		min, _ := strconv.Atoi(split[0])
		max, _ := strconv.Atoi(split[1])
		e := entry{
			min:      min,
			max:      max,
			letter:   split[2],
			password: split[3],
		}
		if strings.Count(e.password, e.letter) <= e.max && strings.Count(e.password, e.letter) >= e.min {
			passwords = append(passwords, e)
		}
	}

	fmt.Println(len(passwords))
}

// Each policy actually describes two positions in the password, where 1 means the first character, 2 means the second character, and so on. (Be careful; Toboggan Corporate Policies have no concept of "index zero"!) Exactly one of these positions must contain the given letter. Other occurrences of the letter are irrelevant for the purposes of policy enforcement.
// Given the same example list from above:
// 1-3 a: abcde is valid: position 1 contains a and position 3 does not.
// 1-3 b: cdefg is invalid: neither position 1 nor position 3 contains b.
// 2-9 c: ccccccccc is invalid: both position 2 and position 9 contain c.
// How many passwords are valid according to the new interpretation of the policies?

func partTwo() {
	f, _ := os.Open("input")
	reader := bufio.NewReader(f)
	scanner := bufio.NewScanner(reader)

	passwords := []entry{}

	for scanner.Scan() {
		split := strings.FieldsFunc(scanner.Text(), fn)
		min, _ := strconv.Atoi(split[0])
		max, _ := strconv.Atoi(split[1])
		e := entry{
			min:      min - 1, // provided as 1-index instead of 0-index!
			max:      max - 1,
			letter:   split[2],
			password: split[3],
		}
		passwordSlice := strings.Split(e.password, "")
		if (passwordSlice[e.min] == e.letter && passwordSlice[e.max] != e.letter) ||
			(passwordSlice[e.min] != e.letter && passwordSlice[e.max] == e.letter) {
			passwords = append(passwords, e)
		}
	}

	fmt.Println(len(passwords))
}
